// Download.swift

import Foundation
import ArgumentParser
import Get

extension Seanut {    
    struct DownloadCommand: AsyncParsableCommand {
        static var configuration = CommandConfiguration(
          commandName: "download",
          abstract: "Downloads media into specified directory (default to current directory)"
        )

        @OptionGroup var options: Seanut.CommonArguments
        
        @Argument(help: "media id returned from search")
        var mediaId: String

        @Option(name: .shortAndLong, help: "specifies that we're only looking for a specific season of a show")
        var season: Int?
        
        @Option(name: .shortAndLong, help: "location to save downloaded media")
        var output: String = FileManager.default.currentDirectoryPath

        static var userId: String!
        static var client: APIClient!
        static var config: RequestConfiguration!
        static let TypesWithChildren: [String] = [
          "MusicAlbum", "MusicGenre", "MusicArtist", "Playlist", "Series", "BoxSet"
        ]
        
        func createDirectory(_ path: String) {
            Seanut.Logger?.info("Creating output directory \(output)...")
            do {
                try FileManager.default.createDirectory(
                  atPath: path.expandingTildeInPath.removingPercentEncoding!,
                  withIntermediateDirectories: true
                )
            } catch {
                fatalError("unable to create output directory at \(path).")
            }
        }

        func downloadRoot(id: String, outputPath: URL) async {
            let downloadInfoReq: Request<Item> = Seanut.jellyfinRequest(
              result: Item(),
              config: DownloadCommand.config,
              path: "/Users/\(DownloadCommand.userId!)/Items/\(id)",
              method: .get
            )

            guard let rootInfo = try? await DownloadCommand.client.send(downloadInfoReq).value else {
                return
            }
            
            if DownloadCommand.TypesWithChildren.contains(rootInfo.type!) || rootInfo.childCount ?? 0 > 0 {
                let newPath = outputPath.appendingPathComponent(rootInfo.name!)
                createDirectory(String(newPath.absoluteString.suffix(newPath.absoluteString.count - 7)))

                let childReq: Request<SearchResult> = Seanut.jellyfinRequest(
                  result: SearchResult(),
                  config: DownloadCommand.config,
                  path: "/Items",
                  method: .get,
                  query: [
                    ("userId", DownloadCommand.userId!),
                    ("parentId", id),
                    ("fields", "Path,ChildCount")
                  ]
                )
                guard let childInfo = try? await DownloadCommand.client.send(childReq).value.items else {
                    return
                }

                Seanut.Logger?.info("Fetching children for \(rootInfo.name!)...")
                
                for child in childInfo {
                    await downloadRoot(id: child.id!, outputPath: newPath)
                }
                
            } else {
                Seanut.Logger?.info("Downloading \(rootInfo.name!)...")
                let pathUrl = URL(fileURLWithPath: rootInfo.path!)
                let itemName: String = rootInfo.name! + "." + pathUrl.pathExtension
                let filePath = outputPath.appendingPathComponent(itemName.addingPercentEncoding(withAllowedCharacters: .urlHostAllowed)!)
                
                await downloadItem(id: id, outputPath: filePath)
            }
        }
        
        func downloadItem(id: String, outputPath: URL) async {
            do {
                let dlReq: Request<Data> = Seanut.jellyfinRequest(
                  result: Data(),
                  config: DownloadCommand.config,
                  path: "/Items/\(id)/Download",
                  method: .get
                )

                #if os(Linux)
                try await DownloadCommand.client.download(for: dlReq, to: outputPath)
                
                #else
                let response = try await DownloadCommand.client.download(for: dlReq)
                try FileManager.default.moveItem(at: response.location, to: outputPath)
                
                #endif
            } catch {
                fatalError("Encountered \(error) downloading media. Please try again later.")
            }
        }
        
        mutating func run() async {
            Seanut.maybeSetLogger(options)
            
            // checks if our specified output directory exists,
            // if it doesn't then we create it before continuing
            if !FileManager.default.fileExists(atPath: output.expandingTildeInPath) {
                createDirectory(output)
            }

            DownloadCommand.config = Seanut.generateJellyfinConfiguration(
              url: options.domain.toURL()!,
              token: Seanut.retrieveAccessToken(for: options.domain.toURL()!)
            )
            
            DownloadCommand.client = APIClient(baseURL: options.domain.toURL()!)

            do {
                let getUserReq: Request<User> = Seanut.jellyfinRequest(
                  result: User(),
                  config: DownloadCommand.config,
                  path: "/Users/Me",
                  method: .get
                )
                
                DownloadCommand.userId = try await DownloadCommand.client.send(getUserReq).value.id!
            } catch {
                fatalError("failed to fetch user id. quitting...")
            }

            await downloadRoot(id: mediaId, outputPath: URL(fileURLWithPath: output))
        }
    }
}
