// The Swift Programming Language
// https://docs.swift.org/swift-book

import Foundation
import ArgumentParser
import Get
import Crypto
import Logging

let SeanutVersion = "0.0.1"

@main
struct Seanut: AsyncParsableCommand {
    static var configuration = CommandConfiguration(
      abstract: "A utility to download media from jellyfin servers",
      subcommands: [SearchCommand.self, DownloadCommand.self, LoginCommand.self],
      defaultSubcommand: SearchCommand.self
    )

    static var Logger: Logger? = nil

    struct CommonArguments: ParsableArguments {
        @Option(name: .shortAndLong, help: "jellyfin server domain name")
        var domain: String

        @Flag(name: .long, help: "prints extra information to the console")
        var verbose = false
    }

    static func maybeSetLogger(_ opts: CommonArguments) {
        Seanut.Logger = opts.verbose ? Logging.Logger(label: "garden.focks.SeanutSwift.logger") : nil
    }

    static func generateJellyfinConfiguration(url: URL, token: String? = nil) -> RequestConfiguration {
        let hostname = Host.current().localizedName ?? "Seanut-Device"
        let digest = SHA256.hash(data: hostname.data(using: .utf8)!)
        
        return RequestConfiguration(
          url: url,
          client: "SeanutSwift",
          deviceName: hostname,
          deviceId: digest.compactMap({ String(format: "%02x", $0) }).joined(),
          version: SeanutVersion,
          token: token
        )
    }

    static func jellyfinRequest<T>(
      result: T,
      config: RequestConfiguration,
      path: String,
      method: HTTPMethod,
      body: Encodable? = nil,
      query: [(String, String?)]? = nil
    ) -> Request<T> {
        let id = String(path.split(separator: "/").last!)
        return Request(
          path: path,
          method: method,
          query: query,
          body: body,
          headers: [ "Authorization": config.authString ],
          id: id
        )
    }

    static func getAccessToken(client: APIClient, username: String, password: String?) async {
        let pass: String? = password ?? {
            print("password>", terminator: " ")
            return readLine(strippingNewline: true)
        }()
        
        let domain = client.configuration.baseURL!
        let configuration = generateJellyfinConfiguration(url: domain)
        let fileName = "~/.seanut/\(domain.host!)".expandingTildeInPath
        
        do {
            let signIn: Request<AuthenticationResponse> = jellyfinRequest(
              result: AuthenticationResponse(),
              config: configuration,
              path: "/Users/AuthenticateByName",
              method: .post,
              body: UserAuthentication(password: pass, username: username)
            )

            let resp = try await client.send(signIn).value
            try resp.accessToken!.write(to: URL(fileURLWithPath: fileName), atomically: false, encoding: .utf8)
           
            print("Access token retrieved ☑️")
        } catch {
            print("Failed to login with provided credentials. please try again later.")
            exit(withError: error)
        }
    }

    static func retrieveAccessToken(for domain: URL) -> String? {
        let tokenPath = FileManager.default
          .homeDirectoryForCurrentUser
          .appendingPathComponent(".seanut/\(domain.host!)")
        return try? String(contentsOf: tokenPath) 
    }
}
